#
# $Copyright: Copyright (c) 2014 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
#
package VRTS::VCS::S3vbs;

require VRTS::Agent;
require VRTS::CMD;
require VRTS::Util;
require VRTS::Paths;
require VRTS::OS;
require VRTS::EXITCODE;
require VRTS::Table;
require VRTS::Summary;
require Sys::Hostname;

use strict;
require Exporter;
use vars qw($VERSION @ISA @EXPORT );
use File::Spec;
use JSON;

@ISA = qw(Exporter);

$VERSION = '5.00';

@EXPORT = qw(initialize discover);

require VRTS::Constants::_HOST_INFO;

require VRTS::Constants::_SF_HOST_TO_HC;
require VRTS::Constants::_SF_HOST_CONTAINER;

use constant TRUE => 1;
use constant FALSE => 0;


# Globals
my $g_hostid =  VRTS::OS->get_hostguid();
my $shortestform_hostname = Sys::Hostname::hostname();
my $devnull = &VRTS::CMD::DEVNULL;
my $g_localclus = " -localclus ";
my $statesfile = "";
my $ddnafile = "";
my $vbs_conf_file = "";
my $res_vbs = "vbsapp";
my $grp_vbs_vcs = "ClusterService";
my $grp_vbs = "VBSHAService";
my $notifylog;
my $logdir;
my ($HASYS);
my $cout;
#Attributes for vbsd
#my @dr_attributes
my $sg_scope = "Sgscope";
my $local_state = "Lstate";
my $remote_clus_name = "RemoteClusterName";
my $remote_clus_state = "RemoteClusterState";
my $remote_sg_state = "RemoteClusterSgState";
my $global_authority = "GlobalGroupAuthority";
#variables
my $parallel = "Parallel";
my $scope_global = "Global";
my $attributes = "Attributes";
my $attname = "AttributeName";
my $attvalue = "Value";
my $vbssvc;

sub initialize
{
	my $agent = shift;
}

sub vbs_log_initialize
{
	$notifylog = new VRTS::Logger(
				Module => 'statenotification',
				Logdir => $logdir,
				Level => 'debug',
				Logfile => 'vbs_state_notify.log',
				UmiString => 'V-0000-0000');
}

sub vbs_log
{
	my $msg  = shift;
	my $sev  = shift;
	if (not defined $notifylog) {
		return;
	}
	if (defined $sev) {
		$notifylog->error($msg);
	} else {
		$notifylog->info($msg);
	}
}

sub discover_vbs_installation
{
	my $agent = shift;
	my $vbs_dir = undef;
	my ($vbs_bin_dir, $vbs_conf_dir) = (undef, undef);
	my $suffix = "";

	if($^O =~ /Win32/i) {
		$vbs_dir = VRTS::OS::querykey("SOFTWARE\\VERITAS\\VRTSvbs\\Base" , "HomeDirectory");
		if (not defined $vbs_dir)
		{
			VRTS::Logger::log('error', "VBS installation does not exist. Exiting...");
			return FALSE;
		}
		$suffix = ".exe";
		$vbs_conf_dir = File::Spec->catpath(undef, $vbs_dir, "conf");
		$logdir = $vbs_dir."\\log";
	} else {
		$vbs_dir = "/opt/VRTSvbs";
		$vbs_conf_dir = "/etc/VRTSvbs";
		$logdir = "/var/VRTSvbs/log/";
	}

	$vbs_bin_dir = File::Spec->catpath(undef, $vbs_dir, "bin");

	$vbssvc = File::Spec->catpath(undef, $vbs_bin_dir, "vbssvc".$suffix);
	$statesfile = File::Spec->catpath(undef, $vbs_conf_dir, ".grp.faults.json");
	$ddnafile = File::Spec->catpath(undef, $vbs_conf_dir, ".grp.node.ddna.json");
	$vbs_conf_file = File::Spec->catpath(undef, $vbs_conf_dir, "vbs.conf");

	if ( ! -f "$vbssvc") {
		VRTS::Logger::log('error', "vbssvc executable $vbssvc does not exist. Exiting...");
		return FALSE;
	}
	
	if($^O =~ /Win32/i) {
		$vbssvc = "\"" . $vbssvc . "\""; 
	}

	return TRUE;
}

sub discover_vcs_installation
{
	my ($vcs_dir, $vcs_bin_dir);
	my $suffix = "";

	if($^O =~ /Win32/i) {
		$vcs_dir = VRTS::OS::querykey("SOFTWARE\\VERITAS\\VCS\\Base" ,"HomeDirectory");
		$suffix = ".exe";
	} else {
		$vcs_dir = $ENV{"VCS_HOME"} || "/opt/VRTSvcs";
	}

	$vcs_bin_dir = File::Spec->catpath(undef, $vcs_dir, "bin");
	$HASYS = File::Spec->catpath(undef, $vcs_bin_dir, "hasys".$suffix);
	if($^O =~ /Win32/i) {
		$HASYS = "\"" . $HASYS . "\""; 
	}
}

sub escape_json_string
{
	my $json = shift;
	
	if($^O =~ /Win32/i)
	{
		# Escaping quotes in the JSON message on Win32
		$json =~  s/"/\\"/g;
		# Adding quotes at the beginning and end
		$json = "\"$json\"";
	}
	else 
	{
		# Single quotes in case of Unix
		$json = "\'$json\'";
	}
	
	return $json;
}

sub update_ddna_file
{
	my ($old_ddna_states, $new_node_fault_ddna) = @_;
	my %final_ddna_hash;
	if ( defined $old_ddna_states) {
		%final_ddna_hash = (%$old_ddna_states,  %$new_node_fault_ddna);
	} else {
		%final_ddna_hash = (%$new_node_fault_ddna);
	}
	if (keys %final_ddna_hash) 
	{
		my $json_ddna = new JSON;
		my $json_data1 = $json_ddna->objToJson(\%final_ddna_hash);
		write_file($ddnafile, $json_data1);
	} else {
		delete_file($ddnafile);
	}
}

my $get_tier_type = sub {
	my $agent = shift;
	my $clus_entry = undef;
	my $plat_type = undef;

	my $clus_tb = $agent->gettable("_SF_HOST_CONTAINER");
	
	if (!$clus_tb) {
		return $plat_type;
	}

	my $clusters = $clus_tb->each_current();
	$clus_entry = $clusters->[0];
	$plat_type = $clus_entry->[1]->
						[&VRTS::Constants::_SF_HOST_CONTAINER::HA_TYPE];

	$plat_type = "MSFOC" if ($plat_type eq "MSCS");

	return $plat_type;
};

sub get_local_details
{
	# my ($tbl_hc, $tbl_hthc) = @_;
	my $tbl_hc = shift;
	my $tbl_hthc = shift;

	my $local_details = {};

	$local_details->{"HOST_ID"} = VRTS::OS->get_hostguid();

	my $host_clus = $tbl_hthc->getrowwithvalue(
									&VRTS::Constants::_SF_HOST_TO_HC::HOST_ID,
									$g_hostid);

	if (defined $host_clus) {
		$local_details->{"NODE_NAME"} =
					$host_clus->[&VRTS::Constants::_SF_HOST_TO_HC::NODE_NAME];
	}

	my $clus_row = $tbl_hc->each_current();
	my $cluster = $clus_row->[0];
	$local_details->{"CLUSTER_ID"} =
					$cluster->[1]->[&VRTS::Constants::_SF_HOST_CONTAINER::ID];
	$local_details->{"CLUSTER_NAME"} =
					$cluster->[1]->[&VRTS::Constants::_SF_HOST_CONTAINER::NAME];

	return $local_details;
}

my %sg_state_to_string = (0 => "OFFLINE",
						  1 => "ONLINE",
						  2 => "PARTIAL",
						  3 => "FAULTED",
						  4 => "PARTIAL|FAULTED",
						  5 => "STARTING",
						  6 => "STOPPING",
						  7 => "FAULTING",
						  8 => "",
						  9 => "",
						  10 => "UNKNOWN");

my %res_state_to_string = (0 => "OFFLINE",
						   1 => "ONLINE",
						   3 => "FAULTED",
						   4 => "STARTING",
						   5 => "STOPPING");

sub get_state_update_json_structure
{
	my $json_data = {};

	$json_data->{"OperationType"} = "-update";
	$json_data->{"ObjectType"} = "ServiceGroup";
	$json_data->{"CLIOperation"} = 1;
	$json_data->{"Objects"} = [];

	return $json_data;
}

sub get_obj_state_json_entry
{
	my $json_data = {};
	my $attr_json = {"AttributeName" => "State"};

	$json_data->{"ObjectName"} = "";
	$json_data->{"Attributes"} = [];
	push(@{$json_data->{"Attributes"}}, $attr_json);

	return $json_data;
}

sub get_res_state_in_grp
{
	my ($res, $grp, $tbl_sg, $tbl_res, $tbl_host_res) = @_;
	my $ret = {};

	# Get matching record from _VCS_SERVICEGROUP
	my $sg_rows = $tbl_sg->getallrowswithvalue(
							&VRTS::Constants::_VCS_SERVICEGROUP::NAME, $grp);
	if (not defined $sg_rows) {
		return $ret;
	}

	my $sg_rec = $sg_rows->[0];
	my $clus_id = $sg_rec->[&VRTS::Constants::_VCS_SERVICEGROUP::HC_ID];
	my $sg_id = $sg_rec->[&VRTS::Constants::_VCS_SERVICEGROUP::ID];

	# Get matching record from _VCS_RESOURCE
	my @cols = (&VRTS::Constants::_VCS_RESOURCE::NAME,
				&VRTS::Constants::_VCS_RESOURCE::SG_ID);
	my @vals = ($res, $sg_id);
	my $res_rec = $tbl_res->getrowwithvaluez(\@cols, \@vals);
	my $res_id = $res_rec->[&VRTS::Constants::_VCS_RESOURCE::ID];

	# Get state info for resource on different hosts
	my $res_rows = $tbl_host_res->getallrowswithvalue(
							&VRTS::Constants::_VCS_HOST_RES_RELATION::RES_ID,
							$res_id);

	if (not defined $res_rows) {
		return $ret;
	}

	for my $res_row (@$res_rows) {
		my $host_id =
			$res_row->[&VRTS::Constants::_VCS_HOST_RES_RELATION::HOST_ID];
		my $res_state =
			$res_row->[&VRTS::Constants::_VCS_HOST_RES_RELATION::STATE];
		$ret->{$host_id} = $res_state_to_string{$res_state};
	}

	return $ret;
}

sub get_all_grp_states_on_host
{
	my ($host_id, $tbl_hst_sg, $grp_states) = @_;

	my $sg_states = $tbl_hst_sg->getallrowswithvalue(
							&VRTS::Constants::_VCS_HOST_SG_RELATION::HOST_ID,
							$host_id);

	if (not defined $sg_states) {
		return $grp_states;
	}

	for my $rowref (@$sg_states) {
		my @row = @$rowref;
		my $sg_id = $row[&VRTS::Constants::_VCS_HOST_SG_RELATION::SG_ID];
		my $state = $row[&VRTS::Constants::_VCS_HOST_SG_RELATION::STATE];
		if (not exists $grp_states->{$sg_id}) {
			$grp_states->{$sg_id} = [];
		}
		push(@{$grp_states->{$sg_id}}, $sg_state_to_string{$state});
	}

	return $grp_states;
}

sub grp_states_to_json
{
	my ($grp_state_hash, $local_details, $sg_lookup) = @_;
	my $json_data = get_state_update_json_structure();

	for my $sg_id (keys %$grp_state_hash) {
		my $sg_states = {};
		my $sg_state_attr = {};
		my $sg_name = $sg_lookup->{$sg_id};
		$sg_states->{"ObjectName"} = $local_details->{"CLUSTER_NAME"} . ":"
																. $sg_name;
		$sg_states->{$attributes} = [];
		$sg_state_attr->{$attname} = "State";
		$sg_state_attr->{$attvalue} = compute_state($grp_state_hash->{$sg_id});
		push(@{$sg_states->{$attributes}}, $sg_state_attr);
		push(@{$json_data->{"Objects"}}, $sg_states);
	}

	return $json_data;
}

sub sgid_to_sg
{
	my $tbl_sg = shift;
	my $cp_tbl_sg = $tbl_sg->copy_current();
	my $ret = {};

	for my $row (@$cp_tbl_sg) {
		my @arow = @$row;
		$ret->{$arow[&VRTS::Constants::_VCS_SERVICEGROUP::ID]}
			 = $arow[&VRTS::Constants::_VCS_SERVICEGROUP::NAME];
	}

	return $ret;
}

sub hostid_to_host
{
	my $tbl_hthc = shift;
	my $cp_tbl_hthc = $tbl_hthc->copy_current();
	my $ret = {};

	for my $row (@$cp_tbl_hthc) {
		my @arow = @$row;
		$ret->{$arow[&VRTS::Constants::_SF_HOST_TO_HC::HOST_ID]}
			 = $arow[&VRTS::Constants::_SF_HOST_TO_HC::NODE_NAME];
	}

	return $ret;
}

sub discover_msfoc
{
	my $agent = shift;

	my $tbl_sg = $agent->gettable("_VCS_SERVICEGROUP");
	my $tbl_hc = $agent->gettable("_SF_HOST_CONTAINER");
	my $tbl_hthc = $agent->gettable("_SF_HOST_TO_HC");
	my $tbl_res = $agent->gettable("_VCS_RESOURCE");
	my $tbl_host_res = $agent->gettable("_VCS_HOST_RES_RELATION");
	my $tbl_hst_sg = $agent->gettable("_VCS_HOST_SG_RELATION");

	my $local_details = get_local_details($tbl_hc, $tbl_hthc);

	# is vbsapp running? on local node?
	my $vbsapp_state = get_res_state_in_grp($res_vbs, $grp_vbs, $tbl_sg,
											$tbl_res, $tbl_host_res);
	my $hstid = $local_details->{"HOST_ID"};
	if (not exists $vbsapp_state->{$hstid}
		|| $vbsapp_state->{$hstid} !~ /ONLINE/) {
		VRTS::Logger::log('debug', "Nothing to report as no vbsapp "
									."resource found");	
		return 1;
	}

	# Get the state of each service group per node.
	my $sg_lookup = sgid_to_sg($tbl_sg);
	my $host_lookup = hostid_to_host($tbl_hthc); # The lookup isn't used yet
	my $grp_state_hash = {};
	my $sg_rows;
	my $json_data = get_state_update_json_structure();

	for my $sg_id (keys %$sg_lookup) {
		my $sg_data = get_obj_state_json_entry();
		$sg_rows = $tbl_sg->getallrowswithvalue(
									&VRTS::Constants::_VCS_SERVICEGROUP::ID,
									$sg_id);

		next if (not defined $sg_rows);
		my $sg_rec = $sg_rows->[0];
		$sg_data->{"ObjectName"} = $local_details->{"CLUSTER_NAME"} . ":"
										. $sg_lookup->{$sg_id};
		my $sg_state = $sg_rec->[&VRTS::Constants::_VCS_SERVICEGROUP::STATE];
		$sg_data->{"Attributes"}->[0]->{"Value"} =
				compute_state([$sg_state_to_string{$sg_state}]);
		push(@{$json_data->{"Objects"}}, $sg_data);
	}

	# generate the json and escape it for use on cmdline
	my $json_str = escape_json_string(to_json($json_data));

	# send the escaped json to vbsd
	my $ret_code = execute_command("$vbssvc -send $json_str");
	if ($ret_code) {
		vbs_log("Failed to send state updates $json_str with error $cout", 1);
	} else {
		vbs_log("Sucessfully sent state updates $json_str");
	}
}


sub get_old_state_for_sg
{
	my $sgs_state_old = shift;
	my $grp = shift;
	if (defined $sgs_state_old) {
		foreach my $row_old(@$sgs_state_old)
		{
			my $grp_itr = $row_old->[&VRTS::Constants::VCS_SG_STATE::NAME];
			if ($grp_itr eq $grp) {
				my $old_state = $row_old->[&VRTS::Constants::VCS_SG_STATE::STATE];
				return $old_state;
			} 
		}
	}
	return undef;

}

sub discover
{
	my $agent = shift;
	my $tier_type = undef;
	my $family = "VCS";

	VRTS::Logger::log('debug', "Entering VBS Discovery");

	$tier_type = $get_tier_type->($agent);
	if (!$tier_type) {
		return undef;
	}

	VRTS::Logger::log('debug', "Using the cluster entries from $family family");

	if (not discover_vbs_installation($agent))
	{
		# Verify this return value
		return undef;
	}

	vbs_log_initialize();

	if ($tier_type eq "VCS") {
		discover_vcs_installation();
		discover_vcs($agent);
	} elsif ($tier_type eq "MSCS" || $tier_type eq "MSFOC") {
		return discover_msfoc($agent);
	}


}
sub discover_vcs
{

	my $agent = shift;
	my %grp_diff_states_hash = ();
	my %group_all_states_hash = ();
	my %group_prev_states_hash = ();
	my %new_fault_bz_node_fault = ();
	my %new_node_fault_ddna = ();

	my $global = $agent->gettable("_VCS_SERVICEGROUP");		
	my $hc_row = $agent->gettable("_SF_HOST_CONTAINER");
	my $host_tb = $agent->gettable("_VCS_NODE_STATUS");
	my $clus_to_host = $agent->gettable("_SF_HOST_TO_HC");
	my $global_state = $agent->gettable("VCS_SG_TO_RHC");		
	my $rhc = $agent->gettable("VCS_HC_TO_RHC");
	my $sg_state_tb = $agent->gettable("VCS_SG_STATE");

	my $local_details = get_local_details($hc_row, $clus_to_host);
	my $node_name = $local_details->{"NODE_NAME"};
	my $cluster_id = $local_details->{"CLUSTER_ID"};
	my $cluster_name = $local_details->{"CLUSTER_NAME"};

	my $entries = $host_tb->copy_current();

	my @hosts_list;

	foreach my $entry (@{$entries})
	{
		my @row = @$entry;
		push (@hosts_list, $row[&VRTS::Constants::_VCS_NODE_STATUS::VCS_NODE_NAME]);
	}
	if (not is_vbsapp_configured($agent)) {
		return;
	}

	my $old_ddna_states = read_file($ddnafile);

	# Hash to store ("SG_name" => "list of nodes on which SG is faulted")
	my %sg_faulted_on_nodes  = ();
	# Hash to store ("SG_name" => No of nodes on which SG is present)
	my %sg_to_node_cnt = ();

	foreach my $hst(@hosts_list)
	{
		my $sg_state_new = $sg_state_tb->getallrowswithvalue(&VRTS::Constants::VCS_SG_STATE::HOST_NAME, $hst);
		my $sgs_state_old = $sg_state_tb->getallrowswithvalue_r(&VRTS::Constants::VCS_SG_STATE::HOST_NAME, $hst);
		if (defined $sg_state_new)
		{
			foreach my $row_new(@$sg_state_new)
			{
				my $grp = $row_new->[&VRTS::Constants::VCS_SG_STATE::NAME];
				my $new_state = $row_new->[&VRTS::Constants::VCS_SG_STATE::STATE];

				# If SG is present on Node increment it's count
				$sg_to_node_cnt{$grp} = $sg_to_node_cnt{$grp} + 1;

				# If SG is faulted, then add it to the faulted node list
				if ($new_state =~ /FAULTED/) {
				    # Append node name to faulted list
				    push(@{$sg_faulted_on_nodes {$grp}}, $hst);
				}

				if ( not exists $group_all_states_hash{$grp}) {
					$group_all_states_hash{ $grp } = [];
				}
				if (defined $group_all_states_hash{$grp}) {
					push (@{$group_all_states_hash{$grp}}, $new_state);
				}
				my $old_state = undef;
				if ($sgs_state_old) {
					$old_state = get_old_state_for_sg($sgs_state_old, $grp);
				}

				if (not defined $old_state) {
					vbs_log("No previous state for service group $grp");
					next;
				}
				if ( not exists $group_prev_states_hash{$grp}) {
					$group_prev_states_hash{ $grp } = [];
				}
				if (defined $group_prev_states_hash{$grp}) {
					push (@{$group_prev_states_hash{$grp}}, $old_state);
				}

				# Old state was FAULTED means we had already sent FAULT 
				# notification. If new state is also FAULTED then there is
				# no need to send FAULT notification again. Also add STATE to
				# grp_diff_states_hash only if NewFault notification is
				# required.
				if(	$new_state =~ m /.*FAULTED.*/i &&
					$old_state !~ m /.*FAULTED.*/i) {
					if ( not exists $grp_diff_states_hash{$grp}) {
						$grp_diff_states_hash{ $grp } = [];
					}
					if (defined $grp_diff_states_hash{$grp}) {
						push (@{$grp_diff_states_hash{$grp}}, $new_state);
					}
				}
				check_if_node_fault_or_ddna($agent, $grp, $old_state, $new_state, $old_ddna_states, 
					\%new_fault_bz_node_fault, \%new_node_fault_ddna, $hst);
			}
		}
	}

	update_ddna_file($old_ddna_states, \%new_node_fault_ddna);
	my ($vbsd_running, $vbsd_online_locally, $host_name) 
					= get_vbsd_local_running_status($agent, $node_name);

	if ($vbsd_running && !$vbsd_online_locally)  {
		VRTS::Logger::log('debug', "Nothing to report as vbsd is running on $host_name");	
		delete_file($statesfile);
		return;
	}
	my $vbsdname = $grp_vbs_vcs;
	if (not exists $group_all_states_hash{$vbsdname}) {
		return;
	}
	my $stored_states;
	my %updated_states = ();
	my $faulttolerancelimit = 0;
	my $vbsdfault = is_grp_new_faulted($vbsdname, \%grp_diff_states_hash, \%new_fault_bz_node_fault);
	my $startlogging = 0;
	if((!$vbsd_online_locally && $vbsdfault) || (-f "$statesfile")) {
		$startlogging = 1;
		$stored_states = read_file($statesfile);
		$faulttolerancelimit = read_value_from_file($vbs_conf_file, "FAULT_RESET_TIMEOUT");
	}

	my $json_data = get_state_update_json_structure();
	foreach my $grpname ( keys %group_all_states_hash )
	{
		my $newfault = 0;
		my $sg_states = {};
		my $sg_state_attr = {};
		my @states = @{$group_all_states_hash{$grpname}};

		$sg_states->{"ObjectName"} = $cluster_name . ":" . $grpname;
		$sg_states->{$attributes} = [];
		$sg_state_attr->{$attname} = "State";
		$sg_state_attr->{$attvalue} = compute_state(\@states);
		push(@{$sg_states->{$attributes}}, $sg_state_attr);
				
		# Store node names on which SG is faulted
		my $faulted_node_list_attr = {};

		# If SG is faulted then add "FaultedNodeList" attribute and insert 
		# node names into FaultedNodeList otherwise assign empty list to 
		# FaultedNodeList (else part)
		if (defined @{$sg_faulted_on_nodes {$grpname}}) {
		    # Create new object for "Attributes" array with 
		    # AttributeName 	: "FaultedNodeList" and
		    # Value				: [list of nodes on which SG is faulted]

		    $faulted_node_list_attr->{$attname} = "FaultedNodeList";

			# faulted node count: Number of nodes on which SG is faulted
			my $fn_cnt = @{$sg_faulted_on_nodes {$grpname}};

			# total node count: Number of nodes on which SG is present
			my $tn_cnt = $sg_to_node_cnt{$grpname};

			# SG is faulted on ALL nodes			
			if ($fn_cnt == $tn_cnt) {
				push(@{$faulted_node_list_attr->{$attvalue}}, "ALL");
			} else {
				foreach (@{$sg_faulted_on_nodes {$grpname}}) {
					push(@{$faulted_node_list_attr->{$attvalue}}, $_);
				}
			}

		    # Insert new object into the "Attributes" array
		    push(@{$sg_states->{$attributes}}, $faulted_node_list_attr);
		} else {
			$faulted_node_list_attr->{$attname} = "FaultedNodeList";
			$faulted_node_list_attr->{$attvalue} = [];
		    # Insert new object into the "Attributes" array
			push(@{$sg_states->{$attributes}}, $faulted_node_list_attr);
		}

		my $grp_details = get_sg_details($agent, $grpname);

		if (!$grp_details->{$parallel}) {
			$newfault = is_grp_new_faulted($grpname, 
					\%grp_diff_states_hash, \%new_fault_bz_node_fault);
		} else {
			my @prev_states_array;
			if ($group_prev_states_hash{$grpname}) {
				@prev_states_array = @{$group_prev_states_hash{$grpname}};
			}
			$newfault = is_grp_new_faulted_for_parallel($grpname, 
					\%grp_diff_states_hash, \%new_fault_bz_node_fault, 
					\@states, \@prev_states_array);
		}
		if ($startlogging) {
			update_states_when_vbsd_down($newfault, $grpname, 
					$stored_states, $faulttolerancelimit, \%updated_states);
		}
		if ($newfault) {
			my $sg_attr = {};
			$sg_attr->{$attname} = "NewFault";
			$sg_attr->{$attvalue} = $newfault;
			push(@{$sg_states->{$attributes}}, $sg_attr);
		}
		add_dr_attributes($sg_states, $grp_details);
		push(@{$json_data->{"Objects"}}, $sg_states);
	}
	my $json_str = escape_json_string(to_json($json_data));
	my $ret_code = execute_command("$vbssvc -send $json_str");
	if ($ret_code) {
		vbs_log("Failed to send state updates $json_str with error $cout", 1);
	} else {
		vbs_log("Sucessfully sent state updates $json_str");
	}
	if ($startlogging && $ret_code) {
		my $json = new JSON;
		my $json_data1 = $json->objToJson(\%$stored_states);
		my $json_data2 = $json->objToJson(\%updated_states);
		if ($json_data1 ne $json_data2) {
			update_states_file($json_data2);
		}
	} 
	if (!$ret_code) {
		delete_file($statesfile);
	}

}

sub add_dr_attributes
{
	my $sg_states = shift;
	my $grp_details = shift;
	my $sg_attr = {};
	if ($grp_details->{$sg_scope} eq $scope_global) {
		$sg_attr->{$attname} = $sg_scope;
		$sg_attr->{$attvalue} = $grp_details->{$sg_scope};
		push(@{$sg_states->{$attributes}}, $sg_attr);
		$sg_attr = {};
		$sg_attr->{$attname} = $local_state;
		$sg_attr->{$attvalue} = $grp_details->{$local_state} + 0;
		push(@{$sg_states->{$attributes}}, $sg_attr);
		$sg_attr = {};
		$sg_attr->{$attname} = $remote_clus_name;
		$sg_attr->{$attvalue} = $grp_details->{$remote_clus_name};
		push(@{$sg_states->{$attributes}}, $sg_attr);
		$sg_attr = {};
		$sg_attr->{$attname} = $remote_clus_state;
		$sg_attr->{$attvalue} = $grp_details->{$remote_clus_state};
		push(@{$sg_states->{$attributes}}, $sg_attr);
		$sg_attr = {};
		$sg_attr->{$attname} = $remote_sg_state;
		$sg_attr->{$attvalue} = $grp_details->{$remote_sg_state} + 0;
		push(@{$sg_states->{$attributes}}, $sg_attr);
		$sg_attr = {};
		$sg_attr->{$attname} = $global_authority;
		$sg_attr->{$attvalue} = $grp_details->{$global_authority} + 0;
		push(@{$sg_states->{$attributes}}, $sg_attr);
	}

}

sub check_if_node_fault_or_ddna
{
	my $agent = shift;
	my $grp = shift;
	my $old_state = shift;
	my $new_state = shift;
	my $old_ddna_states = shift;
	my $new_fault_bz_node_fault = shift;
	my $new_node_fault_ddna = shift;
	my $hst = shift;


	my $host_tb = $agent->gettable("_VCS_NODE_STATUS");
	my $host_state_old = $host_tb->getrowwithvalue_r(&VRTS::Constants::_VCS_NODE_STATUS::VCS_NODE_NAME, $hst);
	my $old_vcs_state = $host_state_old->[&VRTS::Constants::_VCS_NODE_STATUS::VCS_STATE];
	my @host_tbl_cols = (&VRTS::Constants::_VCS_NODE_STATUS::VCS_NODE_NAME);
	my @host_tbl_vals = ($hst);
	my $new_host_row = $host_tb->getrowwithvaluez(\@host_tbl_cols, \@host_tbl_vals);
	my $new_vcs_state = $new_host_row->[&VRTS::Constants::_VCS_NODE_STATUS::VCS_STATE];
	if (($new_vcs_state =~ /RUNNING/) && defined $old_ddna_states && 
			exists $old_ddna_states->{$hst}) {
		delete $old_ddna_states->{$hst};
	} elsif ( defined $old_ddna_states && exists $old_ddna_states->{$hst} && 
				$old_vcs_state  eq $new_vcs_state && $new_vcs_state =~ /FAULTED/) {
		my $out = `$HASYS -value $hst UpDownState 2>&1`;
		chomp($out);
		my $return_code = $?;
		if (!$return_code && $out =~ /Down/) {
			#Consider this as down and append this to the node fault sg list and delete from here:
			my $grps = $old_ddna_states->{$hst};
			foreach my $name (@$grps) { 
				if (not exists $$new_fault_bz_node_fault{$name}) {
					$$new_fault_bz_node_fault{ $name } = 1;
					vbs_log ("Group $name is faulted because node $hst has faulted");
				}
			} 
			delete $old_ddna_states->{$hst};
		}
	}
	if($old_state ne $new_state)
	{
		if ($old_vcs_state ne $new_vcs_state) {
			if ($new_vcs_state =~ /FAULTED/ && $old_state =~ /ONLINE/) {
				my $out = `$HASYS -value $hst UpDownState 2>&1`;
				chomp($out);
				my $return_code = $?;
				if (!$return_code && $out =~ /Down/) {
					if (not exists $$new_fault_bz_node_fault{$grp}) {
						$$new_fault_bz_node_fault{ $grp } = 1;
							vbs_log ("Group $grp is faulted because node $hst has faulted");
					}
				} elsif (!$return_code && $out =~ /Up but/) {
					if ( not exists $$new_node_fault_ddna{$hst}) {
						$$new_node_fault_ddna{ $hst } = [];
					}
					if (defined $$new_node_fault_ddna{$hst}) {
						push (@{$$new_node_fault_ddna{$hst}}, $grp);
					}
				}
			}
		} 
	}

}

sub update_states_when_vbsd_down
{
	my $newfault = shift;
	my $grpname = shift;
	my $stored_states = shift;
	my $faulttolerancelimit = shift;
	my $updated_states = shift;
	my $exist = 0;
	if (!$newfault && defined $stored_states && (exists $stored_states->{$grpname})) {
		$newfault = $stored_states->{$grpname};
		$exist = 1;
	} 
	if ($newfault) {
		my $currenttime = time();
		if ($exist) {
			my $old_fault_time = $stored_states->{"$grpname"."\@\@time"};
			if (defined $faulttolerancelimit && 
					($faulttolerancelimit ne "")
					&& ($currenttime - $old_fault_time > $faulttolerancelimit)) {
				vbs_log("Service group $grpname fault has expired FAULT_TIMEOUT");
				#Delete it	
			} else {
				$$updated_states{$grpname} = $newfault;
				$$updated_states{"$grpname"."\@\@time"} = $old_fault_time;
			}
		} else {
			$$updated_states{$grpname} = $newfault;
			$$updated_states{"$grpname"."\@\@time"} = $currenttime;;
		}
	}
}

sub get_sg_details
{

	#my $global = shift;
	my $agent = shift;
	my $grpname = shift;
	my $grp_details = {};
	my $scope = "Local";
	my $lstate = 0;
	my $cluster_id = "Unknown";
	my $group_id = "Unknown";
	my $g_state = 0;
	my $rhc_name = "Unknown";
	my $rhc_state = "Unknown";
	my $rc_state = 0;
	my $authority = 0;
	my $type = "";

	my $global = $agent->gettable("_VCS_SERVICEGROUP");		
	my $global_state = $agent->gettable("VCS_SG_TO_RHC");		
	my $rhc = $agent->gettable("VCS_HC_TO_RHC");

	my $sg_name = $global->getallrowswithvalue(&VRTS::Constants::_VCS_SERVICEGROUP::NAME,$grpname);
	if (defined $sg_name)
	{
		foreach my $old_row(@$sg_name)
		{
			$scope = $old_row->[&VRTS::Constants::_VCS_SERVICEGROUP::SCOPE];
			$cluster_id = $old_row->[&VRTS::Constants::_VCS_SERVICEGROUP::HC_ID];
			$group_id = $old_row->[&VRTS::Constants::_VCS_SERVICEGROUP::ID];
			$lstate = $old_row->[&VRTS::Constants::_VCS_SERVICEGROUP::STATE];
			$authority = $old_row->[&VRTS::Constants::_VCS_SERVICEGROUP::AUTHORITY];
			$type = $old_row->[&VRTS::Constants::_VCS_SERVICEGROUP::TYPE];
		}
	}
	if ($type eq $parallel) {
		$grp_details->{$parallel} = 1;
	} else {
		$grp_details->{$parallel} = 0;
	}

	if ($scope eq $scope_global) 
	{	
		my $sg_state = $global_state->getallrowswithvalue(&VRTS::Constants::VCS_SG_TO_RHC::SG_ID, $group_id);
		if (defined $sg_state) 
		{
			foreach my $old_row(@$sg_state)
			{
				$g_state = $old_row->[&VRTS::Constants::VCS_SG_TO_RHC::STATE];
			}
		}

		my $rh_state = $rhc->getallrowswithvalue(&VRTS::Constants::VCS_HC_TO_RHC::HC_ID, $cluster_id);
		if (defined $rh_state)
		{
			foreach my $old_row(@$rh_state)
			{
				$rhc_name = $old_row->[&VRTS::Constants::VCS_HC_TO_RHC::RHC_NAME];
				$rc_state = $old_row->[&VRTS::Constants::VCS_HC_TO_RHC::CLUSTER_STATE];
			}
		}

		if ($rc_state eq 1)
		{
			$rhc_state = "ALIVE";
		} else {
			$rhc_state = "DOWN";
		}
		$grp_details->{$local_state} = $lstate;
		$grp_details->{$remote_clus_name} = $rhc_name;
		$grp_details->{$remote_clus_state} = $rhc_state;
		$grp_details->{$remote_sg_state} = $g_state;
		$grp_details->{$global_authority} = $authority;
	}
	$grp_details->{$sg_scope} = $scope;
	return $grp_details;
}

sub get_vbsd_local_running_status
{
	my $agent = shift;
	my $node_name = shift;
	my $res_tb = $agent->gettable("VCS_RES_STATE");
	my @res_tbl_cols = (&VRTS::Constants::VCS_RES_STATE::NAME, &VRTS::Constants::VCS_RES_STATE::STATE);
	my @res_tbl_vals = ($res_vbs, "ONLINE");
	my $res_state = $res_tb->getrowwithvaluez(\@res_tbl_cols, \@res_tbl_vals);
	my $vbsd_online_locally = 0;
	my $vbsd_running = 0;
	my $host_name = "";
	if (defined $res_state)  {
		$vbsd_running = 1;
		$host_name = $res_state->[&VRTS::Constants::VCS_RES_STATE::HOST_NAME];
		if ($node_name eq $host_name) {
			$vbsd_online_locally = 1;
		} 
	}
	return ($vbsd_running, $vbsd_online_locally, $host_name);

}

sub is_vbsapp_configured
{
	my $agent = shift;	
	#check if the vbsapp resource is online or not:
	my $res_tb = $agent->gettable("VCS_RES_STATE");
	my $res_state_tb = $res_tb->getrowwithvalue(&VRTS::Constants::VCS_RES_STATE::NAME, $res_vbs);
	if (not defined $res_state_tb) {
		VRTS::Logger::log('debug', "Nothing to report as no vbsapp resource found");	
		return FALSE;
	}
	return TRUE;
}

sub compute_state
{
	my ($aref) = @_;
	my ($totalstate) = 0;
	foreach my $state (@$aref) {
		if ($state =~ m/.*FAULTED.*/i) {
			# To handle FAULTED|PARTIAL state.
			if ($state =~ m/.*PARTIAL.*/i) {
				$totalstate |= 0x008
			}
			$totalstate |= 0x002;
		} elsif ($state =~ m/.*ONLINE.*/i) {
			if ($totalstate & 0x008) {
				$totalstate = $totalstate & ~0x008;
			}
			$totalstate |= 0x004;
		} elsif ($state =~ /.*PARTIAL.*/i && !($totalstate & 0x004)) {
			$totalstate |= 0x008;
		}
	}
	if (($totalstate == 0x002) || ($totalstate == 0)) {
		$totalstate |= 0x001 #offline;
	}
	return $totalstate;
}



sub is_online
{
	my ($state) = @_;
	my $online = 0;
	if ($state & 0x004) {
		$online = 1;
	}
	return $online;

}

sub delete_file
{
	my $file = shift;
	if (-f $file) {
		unlink($file);
	}
}

sub update_states_file
{
	my ($json) = @_;
	my $json_path = File::Spec->catfile($statesfile);
	delete_file($statesfile);#Check if this needed
	write_file($json_path, $json);
}

sub read_file
{
	my $file = shift;
	my $text;
	if (! -f $file) {
		return undef;
	}
	# Read the input file.
	open (FILE, "<".$file) || return undef;
	while(<FILE>) {
		$text .= $_;
	}
	close(FILE);
	my $json = new JSON;
	my $json_hash = $json->jsonToObj ($text);
	return $json_hash;

}


sub write_file
{
    my $json_path = shift;
    my $json_data = shift;

    $json_path = untaintmeta($json_path);

    open(JSONFILE, ">$json_path");
    print JSONFILE $json_data;
    close(JSONFILE);
}

sub read_value_from_file
{
    my (${file}, ${attribute})  = @_;
    my (@tags) = ();
    my ($name, $value, $input) = ("", "", "");
    my @inputs = ();
    open(INPUT1, "${file}") || die("Could not open $file!");
    @inputs = <INPUT1>;
    close (INPUT1);
    foreach $input (@inputs) {
        chomp ($input);

        # remove any beginning or trailing space and squeeze it.
        $input =~ s/\s+/ /g;
        $input =~ s/^\s+//;
        $input =~ s/\s+$//;
        if($input =~ /^${attribute}.*/) {
            @tags = split (/[=#;]/, $input);
            $name = $tags[0]; $name =~ s/^\s//;
            $value = $tags[1]; $value =~ s/^\s//;
            return ($value);
        }
    }
    return ("");
}

sub untaintmeta
{
    my $tainted = shift;
    return undef if(not defined $tainted);
    $tainted =~  m/([A-Za-z0-9\-\_\\\/\:\.\s]+)/;
    return $1;
}


sub is_faulted
{
	my ($aref) = @_;
	my $faulted = 0;
	foreach my $state (@$aref) {
		if ($state =~ m/.*FAULTED.*/i) {
			$faulted = 1;
			last;
		}
	}
	return $faulted;
}


sub is_grp_new_faulted
{
	my ($grpname, $states_hash, $node_fault_states) = @_;
	my $fault = 0;
	if (exists $$node_fault_states{$grpname}) {	
		$fault = 1;
	} elsif (exists $$states_hash{$grpname}) {
		my @array = $$states_hash{$grpname};
		if (scalar(@array) > 0) {
			$fault = 1;
		}
	}
	return $fault;
}


sub is_grp_new_faulted_for_parallel
{
	my ($grpname, $states_hash, $node_fault_states, $new_states, $old_states) = @_;
	my $fault = 0;
	if (exists $$node_fault_states{$grpname} || exists $$states_hash{$grpname}) {
		my $old_states_len = scalar(@$old_states);
		my $new_states_len = scalar(@$new_states);
		my $no_of_fault = 0;
		my $no_of_online = 0;
		
		if ($old_states_len != $new_states_len) {
			return 0;
		}
		for (my $i = 0; $i < $old_states_len; $i++) {
			my $old_state = $old_states->[$i];
			my $new_state = $new_states->[$i];
			if ($old_state eq $new_state && $new_state =~ /ONLINE/) {
				$no_of_online++;
				last;
			}
			if (($old_state =~ /ONLINE/ || $old_state =~ /PARTIAL/) 
						&& $new_state =~ /FAULTED/) {
				$no_of_fault++;
			}
		}
		if ($no_of_online) { #Order is important
			$fault = 0;
		} elsif ($no_of_fault) {
			$fault = 1;
		} elsif (exists $$node_fault_states{$grpname}) {
			$fault = 1;
		}
	}
	return $fault;
}

sub execute_command
{
	my $cmd = shift;
	my $output = shift;
	$cout = `$cmd 2>&1`;
	my $ret = 0;
	chomp($cout);
	if($? != 0)
	{
		if ($output) {
			#VRTS::Logger::log('info',"command $cmd failed: $cout");
			vbs_log("command $cmd failed: $cout", 1);
		}
		$ret = 1;
	}
	else
	{
		if ($output) {
			vbs_log("executed $cmd sucessfully with output:$cout");
			#VRTS::Logger::log('info',"executed $cmd sucessfully with output:$cout");
		}
		$ret = 0;
	}
	return $ret;
}

#define GROUP_OFFLINE		0x001
#define GROUP_FAULTED		0x002
#define GROUP_ONLINE		0x004
#define GROUP_PARTIAL		0x008
#define GROUP_STARTING		0x010
#define GROUP_STOPPING		0x020
#define GROUP_FAILED_START	0x040
#define GROUP_FAILED_STOP	0x080
#define GROUP_PENDING		0x100



1;
