#
# $Copyright: Copyright (c) 2014 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
# $Id: Common.pm,v 1.14 2013/08/29 04:53:04 spatanka Exp $
#
package VBSTier::Common;

use strict;
use Exporter ();
use Carp;
use File::Spec;

our @ISA = qw(Exporter);
our @EXPORT = qw(parse_cmdline run_cmd parse_options dispatch_vbs_status init_log fail_msg
warn_msg debug_msg assert_instance assert_class parse_vbs_conf discover_vbs_installation
logdbg vbs_log_msg);

our $dbg = "";
our $logger = undef;

sub assert_instance {
	my ($s) = @_;

	fail_msg("Cannot call method ".(caller(1))[3].
		" without an instance of ". $s) unless ref $s;
}

sub assert_class {
	my ($s) = @_;

	fail_msg("Cannot call class method ".(caller(1))[3].
		" on an instance") if ref $s;
}

sub init_log {
	my ($log_name, $log_dir) = @_;
	$logger = new VRTS::Logger(Module => $log_name,
							   Logdir => File::Spec->catpath(undef,
					   							$log_dir, "log"),
							   Level => 'debug',
							   Logfile => $log_name . '.log',
							   UmiString => '');
}

sub logdbg {
	my ($dbg_level) = @_;
	if ($dbg_level) {
		$dbg = "1";
	} else {
		$dbg = "";
	}
}

sub fail_msg {
	my ($msg) = @_;

	if (defined $msg) {
		if (defined $logger) {
			$logger->error($msg);
		}
		croak $msg;
	}
}

sub warn_msg {
	my ($msg) = @_;

	if (defined $msg) {
		if (defined $logger) {
			$logger->warning($msg);
		}
		carp $msg;
	}
}

sub debug_msg {
	my ($msg) = @_;

	if ($dbg and defined $msg
			 and defined $logger) {
		my $calling_func = (caller(1))[3];
		$msg = "$calling_func : $msg";
		$logger->debug($msg);
	}
}

# The following is a stop-gap arrangement to fix the case
# where VOM sends ClusterType as MSCS but we need MSFOC.
my $map_clus_plat = sub {
	my ($clus_plat) = @_;

	if ($clus_plat =~ /MSCS/i) {
		$clus_plat = "MSFOC";
	}

	return $clus_plat;
};

sub parse_cmdline {
	my ($cmdopts_ref) = @_;
	use Getopt::Long qw(:config auto_help);
	my %option_hash = ();

	my $result = GetOptions(\%option_hash,
								@$cmdopts_ref);

	if (exists $option_hash{"options"}) {
		my $option_string = $option_hash{"options"};
		if (length($option_string) > 0) {
			my $options = parse_options($option_string);
			$option_hash{"options"} = $options;
		}
	}

	$option_hash{"tier-type"} = $map_clus_plat->($option_hash{"tier-type"});

	return \%option_hash;
}

sub run_cmd {
	my ($cmd_string) = @_;
	my $ret = {};

	fail_msg("Invalid command string in ".(caller(0))[3])
		if not defined $cmd_string;

	use IPC::Open3;
	use IO::Select;
	use Symbol qw(gensym);

	$ret->{"cmd"} = $cmd_string;
	$ret->{"cmd"} =~ s/[<>]//g if ($^O =~ /Win32/i); # Windows cmd-line barfs
													 # on embedded arrows
	$ret->{"code"} = 0;
	$ret->{"out"} = "";
	$ret->{"err"} = "";

	my ($out_str, $err_str) = ("", "");
	my $pid;

	debug_msg("Running: $cmd_string");

	if ($^O !~ /Win32/i) {
		my ($cmd_out, $cmd_err) = map gensym, 1..2;

		$pid = open3(\*STDIN, $cmd_out, $cmd_err, $cmd_string)
				or fail_msg("Failed to run command \'$cmd_string\': $!");

		my $selector = IO::Select->new();
		$selector->add($cmd_out);
		$selector->add($cmd_err);

		while (my @ready = $selector->can_read()) {
			for my $fh (@ready) {
				$out_str .= scalar <$cmd_out> if ($fh == $cmd_out);
				$err_str .= scalar <$cmd_err> if ($fh == $cmd_err);
				if (eof($fh)) {
					$selector->remove($fh);
					close $fh;
				}
			}
		}
	} else {
		my $line;

		$pid = open3(\*STDIN, \*OSTRM, \*ESTRM, $cmd_string) 
					or fail_msg("Failed to execute command $cmd_string");

		my @lines = <OSTRM>;
		for (@lines) {
			s/\r//;
			$out_str .= $_;
		}

		@lines = <ESTRM>;
		for (@lines) {
			s/\r//;
			$err_str .= $_;
		}

		close OSTRM;
		close ESTRM;	
	}

	$ret->{"out"} = $out_str;
	$ret->{"err"} = $err_str;

	waitpid($pid, 0);
	
	$ret->{"code"} = $?;

	debug_msg("Command output: '$out_str'");
	debug_msg("Command error: '$err_str'");
	debug_msg("Command exit code: " . $ret->{"code"});

	return $ret;
}

sub parse_options {
	my ($option_string) = @_;

	# printf "option string = %s\n", $option_string;

	my $options = {};
	my @kv_pairs = split(/,/, $option_string);
	my ($key, $value);

	for my $kv (@kv_pairs) {
		# printf "%s\n", $kv;

		($key, $value) = split(/=/, $kv, 2);
		if ($value) {
			$options->{$key} = $value;
		} else {
			$options->{$key} = undef;
		}
	}

	return $options;
}

my $escape_json_string = sub {
	my ($json) = @_;

	if (not defined $json) {
		return "{}";
	} if (length($json) <= 0) {
		return $json;
	}
	
	if($^O =~ /Win32/i) {
		# Escaping quotes in the JSON message on Win32
		$json =~  s/"/\\"/g;
		# Adding quotes at the beginning and end
		$json = "\"$json\"";
	} else {
		# Quote embedded single quotes 
		#   like this ' -> ' "'" '
		$json =~ s/\'/\'\"\'\"\'/g;

		# Single quotes in case of Unix
		$json = "\'$json\'";
	}

	# Replace embedded new-lines
	#   with spaces.
	$json =~ s/\n/ /g;

	# Zap embedded carriage-returns
	$json =~ s/\r//g;
	
	return $json;
};

my $escape_quoted_cmd = sub {
	my ($cmd) = @_;

	# Windows needs an extra
	# backslash for escaping
	if ($^O =~ /Win32/i) {
		$cmd =~ s/"/\\\\"/g;
	} else {
		$cmd =~ s/"/\\"/g;
	}

	return $cmd;
};

my $replace_quotes_with_single = sub {
	my ($cmd) = @_;
	$cmd =~ s/\\/\\\\/g; #Fix this 
	$cmd =~ s/"/'/g;
	return $cmd;
};

my $json_response_header = sub {
	my ($operation_type, $object_type, $obj_name) = @_;

	my $json = "{";
	$json .= "\"OperationType\" : \"$operation_type\", ";
	$json .= "\"ObjectType\" : \"$object_type\", ";
	$json .= "\"CLIOperation\" : true, ";
	$json .= "\"Objects\" : [ ";
	$json .= " { ";
	$json .= "\"ObjectName\" : \"$obj_name\", ";
	$json .= "\"Attributes\" : [ ";

	return $json;
};

my $json_response_trailer = sub {
	my $json = "";

	$json .= " ] ";
	$json .= " } ";
	$json .= " ] ";
	$json .= "}";

	return $json;
};

my $json_response = sub {
	my ($resp, $obj) = @_;
	my ($cmd, $exit_code, $out, $err) = ("", 0, "", "");
	my ($taskid, $subtaskid, $source_clus, $obj_name, $vm_state)
				= ("", "", "", "", 0);

	my $operation_type = "-update-subtasks";
	my $object_type = "Task";

	if (defined $resp->{"optype"}) {
		$operation_type	= $resp->{"optype"};
	}

	if (defined $resp->{"objtype"}) {
		$object_type = $resp->{"objtype"};
	}

	if (defined $resp) {
		if (defined $resp->{"out"}) {
			$out = $resp->{"out"};
		}

		if (defined $resp->{"err"}) {
			$err = $resp->{"err"};
		}

		if (defined $resp->{"code"}) {
			$exit_code = $resp->{"code"};
		}

		if (defined $resp->{"cmd"}) {
			$cmd = $resp->{"cmd"};
		}

		if (defined $resp->{"optype"}) {
			my $tmp = $resp->{"optype"};
			if (length($tmp) > 0) {
				$operation_type = $tmp;
			}
		}

		if (defined $resp->{"objtype"}) {
			my $tmp = $resp->{"objtype"};
			if (length($tmp) > 0) {
				$object_type = $tmp;
			}
		}

		if (defined $resp->{"objname"}) {
			my $tmp = $resp->{"objname"};
			if (length($tmp) > 0) {
				$obj_name = $tmp;
			}
		}

		if (defined $resp->{"VMState"}) {
			$vm_state = $resp->{"VMState"};
		}
	}

	if (defined $obj) {
		if (defined $obj->{"task-id"}) {
			$taskid = $obj->{"task-id"};
			if (not defined $obj_name or length($obj_name) <= 0) {
				$obj_name = $taskid;
			}
		}

		if (defined $obj->{"subtask-id"}) {
			$subtaskid = $obj->{"subtask-id"};
		}

		if (defined $obj->{"source-cluster"}) {
			$source_clus = $obj->{"source-cluster"};
		}
	}

	my $json = "";

	if (defined $taskid && length($taskid) > 0
		&& $operation_type eq "-update-subtasks") {
		$json = $json_response_header->($operation_type, $object_type,
										$obj_name);

		$json .= " { \"AttributeName\" : \"SubtaskId\", \"Value\" : $subtaskid }";
		$json .= ", { \"AttributeName\" : \"SourceCluster\", "
						 . "\"Value\" : \"$source_clus\" }";

		if (defined $cmd and length($cmd) > 0) {
			my $qcmd = $replace_quotes_with_single->($cmd);
			$json .= ", { \"AttributeName\" : \"Command\", \"Value\" : \"$qcmd\" }";
		}

		if (defined $err and length($err) > 0) {
			$json .= ", { \"AttributeName\" : \"ErrStr\", \"Value\" : \"$err\" }";
		}

		if (defined $out and length($out) > 0) {
			$json .= ", { \"AttributeName\" : \"OutStr\", \"Value\" : \"$out\" }";
		}

		$json .= ", { \"AttributeName\" : \"ExitCode\", \"Value\" : $exit_code }";
		$json .= $json_response_trailer->();
	} elsif ($operation_type eq "-update") {
		$json = $json_response_header->($operation_type, $object_type,
										$obj_name);
		if (defined $vm_state) {
			$json .= "{ \"AttributeName\" : \"VMState\", \"Value\" : $vm_state }";
		}
		$json .= $json_response_trailer->();
	}

	return $escape_json_string->($json);
};

sub dispatch_vbs_status {
	my ($ret, $obj) = @_;

	my $json = $json_response->($ret, $obj);

	if (length($json) <= 0) {
		return 0;
	}

	my $vbs_path = (discover_vbs_installation())[0];
	my $cmd = File::Spec->catpath('', $vbs_path, "vbssvc");

	$cmd = "\"$cmd\" -send $json";

	my $send_ret = run_cmd($cmd);

	if (defined $send_ret->{"code"}) {
		return $send_ret->{"code"};
	}

	return 0;
}

# Returns an array with the following elements:
# 0: bin directory, 1: conf directory, 2: var directory
# Don't change the order.
sub discover_vbs_installation {
	my @vbs_dirs = ();

	if ($^O !~ /Win32/i) {
		push(@vbs_dirs, "/opt/VRTSvbs/bin");
		push(@vbs_dirs, "/etc/VRTSvbs");
		push(@vbs_dirs, "/var/VRTSvbs");
	} else {
		use VRTS::OS;

		my $dir = VRTS::OS::querykey("SOFTWARE\\VERITAS\\VRTSvbs\\Base",
								"HomeDirectory");

		if (not defined $dir) {
			fail_msg("Virtual Business Services not installed on local machine");
		}

		push(@vbs_dirs, File::Spec->catpath(undef, $dir, "bin"));
		push(@vbs_dirs, File::Spec->catpath(undef, $dir, "conf"));
		push(@vbs_dirs, $dir);
	}

	return @vbs_dirs;
};

sub parse_vbs_conf {
	my (undef, $vbs_conf_dir) = discover_vbs_installation();
	my ($file)  = File::Spec->catpath('', $vbs_conf_dir, "vbs.conf");
	my (@tags) = ();
	my ($name, $value) = ("", "");
	my @inputs = ();
	my $conf = {};

	open(INPUT1, "$file") || fail_msg("Could not open $file");
	@inputs = <INPUT1>;
	close (INPUT1);

	for my $input (@inputs) {
		chomp ($input);

		# skip a comment line
		next if $input =~ /^\s*#/;

		@tags = split (/[=#;]/, $input);
		$name = $tags[0]; $name =~ s/^\s+//; $name =~ s/\s+$//;
		$value = $tags[1]; $value =~ s/^\s//; $value =~ s/\s+$//;
		$conf->{$name} = $value;
	}

	return $conf;
}

# all perl scripts must use this function if it is not a debug msg.
# Derived from VCSAG_LOG_MSG
# usage:
# 1. vbs_log_msg ("<sev>", "<msg>", <msgid>)
# 2. vbs_log_msg ("<sev>", "<msg>", <msgid>, "-vbs <vbs_name>")
# 3. vbs_log_msg ("<sev>", "<msg>", <msgid>, "$param1")
# 4. vbs_log_msg ("<sev>", "<msg>", <msgid>, "-vbs <vbs_name>", "$param1")
# 5. vbs_log_msg ("<sev>", "<msg>", <msgid>, "-encoding <encoding>")
# 6. vbs_log_msg ("<sev>", "<msg>", <msgid>, "-vbs <vbs_name>", "-encoding <encoding>")
# 7. vbs_log_msg ("<sev>", "<msg>", <msgid>, "-encoding <encoding>", "$param1")
# 8. vbs_log_msg ("<sev>", "<msg>", <msgid>, "-vbs <vbs_name>", "-encoding <encoding>", "$param1")
# Note1: Only 6 parameters supported.
# Note2: The "<msg>" must NOT span across multiple lines.
#
# examples:
# 1. vbs_log_msg ("N", "a test message", 1);
# 2. vbs_log_msg ("N", "a test message", 1, "-vbs sales_vbs");
# 3. vbs_log_msg ("N", "a test message of $test", 1, "$test");
# 4. vbs_log_msg ("N", "a test message of $test", 1, "-vbs sales_vbs", "$test");
# 5. vbs_log_msg ("N", "a test message", 1, "-encoding en_US.UTF-8");
# 6. vbs_log_msg ("N", "a test message", 1, "-vbs sales_vbs", "-encoding en_US.UTF-8");
# 7. vbs_log_msg ("N", "a test message of $test", 1, "-encoding en_US.UTF-8", "$test");
# 8. vbs_log_msg ("N", "a test message of $test", 1, "-vbs sales_vbs", "-encoding en_US.UTF-8", "$test");
sub vbs_log_msg
{
	# @params includes "-encoding <encoding>" and all the parameters.
	# caller needs to make sure that it has passed the parameters in
	# the correct order.
	# if <encoding> is passed as an argument then the format of the
	# params would be :
	# $params[0] = "-encoding <encoding>"
	# $params[1] = <param1>
	# $params[2] = <param2> and so on.
	# otherwise :
	# $params[0] = <param1>
	# $params[1] = <param2> and so on.
	my ($VCS_SUPPLANG, $VCS_USERLANG, $VCS_IS_CLANG_SET) = ("C", $ENV{"LANG"}, 1);

	my $vbs_path = (discover_vbs_installation())[0];
	my $VBSLOG = File::Spec->catpath('', $vbs_path, "vbslog");
	$VBSLOG = "\"".$VBSLOG.".exe\"" if ($^O =~ /Win32/i);

	my ($severity, $msg, $msgid, @params) = @_;
	my ($i, $log_msg) = (0, "");
	my ($encoding, $vbs, $param_idx, $parameters) = ("", "", 0, "", "");

	$msg =~ s/"/\\"/g;

	for ($i = 0; ($i < 2) && ($#params >= 0); $i++) { 
		if ($params[0] =~ /^\s*-encoding\s/) {
			$encoding = $params[0];
			shift(@params);
		} elsif ($params[0] =~ /^\s*-vbs\s/) {
			$vbs = $params[0];
			shift(@params);
		} else {
			last;
		}
	}

	for ($i = $param_idx; $i <= $#params; $i++) {
		$params[$i] = "%$params[$i]" if ($params[$i] =~ /^-/);
		$parameters .= "\"$params[$i]\" ";
	}

	$parameters = "-parameters " . "$parameters" if ("$parameters");
	$log_msg = "$VBSLOG -add \"$msg\" -sev $severity $vbs -msgid $msgid";
	$log_msg .= " $encoding $parameters";

	$ENV{"LANG"} = $VCS_USERLANG if ($VCS_IS_CLANG_SET);
#	print("$log_msg\n");
	system ("$log_msg");
	$ENV{"LANG"} = $VCS_SUPPLANG if ($VCS_IS_CLANG_SET);
}

1;

