#
# $Copyright: Copyright (c) 2014 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
# $Id: VCS.pm,v 1.8 2013/04/17 06:40:07 amukher1 Exp $
#
package VBSTier::VCS;

use strict;
use Exporter ();
use Carp;
use VBSTier::Common;

our @ISA = qw(Exporter);
our @EXPORT = qw(new run_operation);

use constant ANY_SYS => "-any";

my $discover_installation = sub {
	my @vcs_dirs = ();

	if ($^O !~ /Win32/i) {
		push(@vcs_dirs, "/opt/VRTSvcs/bin");
		push(@vcs_dirs, "/etc/VRTSvcs");
		push(@vcs_dirs, "/var/VRTSvcs");
	} else {
		use VRTS::OS;

		my $dir = VRTS::OS::querykey("SOFTWARE\\VERITAS\\VCS\\Base",
								"HomeDirectory");

		if (not defined $dir) {
			fail_msg("Veritas Cluster Server not installed on local machine");
		}

		push(@vcs_dirs, File::Spec->catpath(undef, $dir, "bin"));
		push(@vcs_dirs, File::Spec->catpath(undef, $dir, "conf"));
		push(@vcs_dirs, $dir);
	}

	return @vcs_dirs;
};

my $get_ha_cmd = sub {
	my ($sufx) = @_;

	my $cmd = File::Spec->catpath('', ($discover_installation->())[0], "ha".$sufx);
	return "\"$cmd\"";
};

my $get_grp_cmd = sub {
	return $get_ha_cmd->("grp");
};

my $log_to_halog = sub {
	my ($cmd) = @_;
	my $prefix =
		"Symantec Virtual Business Services invoked the following command: ";
	my $payload = $prefix . $cmd;
	$payload =~ s/\"/\\\"/g;

	my $halog_cmd = $get_ha_cmd->("log") . " -add \"$payload\" -sev I";
	my $ret = run_cmd($halog_cmd);
};

sub new {
	my ($type, $self) = @_;
	assert_class($type, $self);

	fail_msg("Cannot instantiate ". $type .
		" on an undefined instance") if (not defined $self);

	return bless($self, $type);
}

sub run_operation {
	my ($self) = @_;
	assert_instance($self);
	my $action = lc($self->{"action"});

	if ($action eq "online") {
		return $self->online();
	} elsif ($action eq "offline") {
		return $self->offline();
	} elsif ($action eq "flush") {
		return $self->flush();
	} elsif ($action eq "query") {
		return $self->query();
	} elsif ($action eq "switch") {
		return $self->switchsg();
    }
}

sub switchsg {
	my ($self) = @_;
	assert_instance($self);

	my $grp_name = $self->{"object-name"};

	my $cmd = $get_grp_cmd->() . " -switch $grp_name";
	$cmd .= "  -any -localclus";

	my $ret = run_cmd($cmd);
	$log_to_halog->($cmd);
	return dispatch_vbs_status($ret, $self);
}

sub online {
	my ($self) = @_;
	assert_instance($self);

	my $grp_name = $self->{"object-name"};
	my $options = $self->{"options"};

	my $cmd = $get_grp_cmd->() . " -online";

	if (defined $options and defined $options->{"takeover"}) {
		if ($options->{"takeover"} eq "1") {
			$cmd .= " -force";
		}
	}

	$cmd .= " $grp_name";

	if (defined $options and defined $options->{"targetsys"}) {
		if ($options->{"targetsys"} eq ANY_SYS) {
			$cmd .= " -any";
		} else {
			$cmd .= " -sys " . $options->{"targetsys"};
		}
	} else {
		$cmd .= "  -any";
	}

	my $ret = run_cmd($cmd);
	$log_to_halog->($cmd);
	return dispatch_vbs_status($ret, $self);
}

sub offline {
	my ($self) = @_;
	assert_instance($self);

	my $grp_name = $self->{"object-name"};
	my $options = $self->{"options"};

	my $cmd = $get_grp_cmd->() . " -offline $grp_name";

	if (defined $options and defined $options->{"targetsys"}) {
		$cmd .= " -any"
			if ($options->{"targetsys"} eq ANY_SYS);
	} else {
		$cmd .= "  -any";
	}

	my $ret = run_cmd($cmd);
	$log_to_halog->($cmd);
	return dispatch_vbs_status($ret, $self);
}

sub flush {
	my ($self) = @_;
	assert_instance($self);

	my $grp_name = $self->{"object-name"};

	my $cmd = $get_grp_cmd->() . " -flush $grp_name";

	my $ret = run_cmd($cmd);
	$log_to_halog->($cmd);
	return dispatch_vbs_status($ret, $self);
}

sub query {
	my ($self) = @_;
	assert_instance($self);
	my $cmd_sufx = "";
	my $obj_type = "";
	my $attr_name = $self->{"attribute"};

	if (not defined $attr_name or length($attr_name) <= 0) {
		fail_msg("Invalid attributes queried");
	}

	if (defined $self->{"object-type"}) {
		$obj_type = $self->{"object-type"};

		if (lc($obj_type) eq "group") {
			$cmd_sufx = "grp";
		} elsif (lc($obj_type) eq "cluster"
						or length($obj_type) <= 0) {
			$cmd_sufx = "clus";
		}
	} else {
		$cmd_sufx = "clus";
	}

	my $obj_name = $self->{"object-name"}; # blank for clus
	my $options = $self->{"options"};

	my $cmd = $get_ha_cmd->($cmd_sufx) . " -value $obj_name $attr_name";

	my $ret = run_cmd($cmd);

	# implicitly assumed to be sync
	print STDOUT $ret->{"out"};
	print STDERR $ret->{"err"};
	
	return $ret->{"code"};
}

1;

