#
# $Copyright: Copyright (c) 2014 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
# $Id: VMware.pm,v 1.5 2013/05/30 09:28:44 amukher1 Exp $
#
package VBSTier::VMware;

use strict;

use Exporter ();
use Carp;
use VBSTier::Common;
use File::Spec;
use VRTS::Spoke;

our @ISA = qw(Exporter);
our @EXPORT = qw(new run_operation);

my $query_vm_id = sub {
	my ($self, $vm_name) = @_;

	my $vbs_path = discover_vbs_installation();
	my $cmd = File::Spec->catpath('', $vbs_path, "vbssvc");
	$cmd = "$cmd -vminfo -details";

	my $send_ret = run_cmd($cmd);

	my @outlines = split(/\n/, $send_ret->{"out"});
	for my $line (@outlines) {
		next if ($line =~ m/Name\s+/);
		my @elements = split(/\s+/, $line);
		if (@elements < 5) {
			# TODO: FIXME: log
			return ("", "");
		}
		if ($vm_name eq $elements[3]) {
			# [4] = vmid, [2] = sgid
			return ($elements[4], $elements[2]);
		}
	}

	return ("", "");
};


sub new {
	my ($type, $self) = @_;
	assert_class($type, $self);

	fail_msg("Cannot instantiate ". $type .
		" on an undefined instance") if (not defined $self);

	return bless($self, $type);
}

sub run_operation {
	my ($self) = @_;
	assert_instance($self);
	my $action = lc($self->{"action"});
	my $vm_name;

	my $vbs_conf = parse_vbs_conf();
	if (not defined $vbs_conf->{"CONFIGURATOR_IPS"}
			or length($vbs_conf->{"CONFIGURATOR_IPS"}) <= 0) {
		fail_msg("Configurator IP addresses are not available."
					. " Aborting VM operation");
	} else {
		$self->{"ConfiguratorIP"} =
					(split(/\s/, $vbs_conf->{"CONFIGURATOR_IPS"}))[0];
	}

	my $options = $self->{"options"};

	if (not defined $options
			or not defined $options->{"vmid"}
			or not defined $options->{"sgid"}) {
		$vm_name = $self->{"object-name"};
		if (not defined $vm_name or length($vm_name) <= 0) {
			fail_msg("Cannot determine vmid, sgid or VM name");
		}

		my ($vm_id, $sg_id) = $query_vm_id->($self, $vm_name);

		if (length($vm_id) <= 0 or length($sg_id) <= 0) {
			fail_msg("Cannot determine mandatory arguments vmid or sgid");
		} else {
			$options->{"vmid"} = $vm_id;
			$options->{"sgid"} = $sg_id;
		}
	}

	if ($action eq "online") {
		return $self->start_vm();
	} elsif ($action eq "offline") {
		return $self->stop_vm();
	} elsif ($action eq "query") {
		return $self->query();
	}
}

sub start_vm {
	my ($self) = @_;
	my $resp = {};
	assert_instance($self);

	my $vm_details = {};
	my $vm_name = $self->{"object-name"};
	my $options = $self->{"options"};
	my $cs_ip = $self->{"ConfiguratorIP"};

	$vm_details->{"VM_ID"} = $options->{"vmid"};
	$vm_details->{"SG_ID"} = $options->{"sgid"};
	$vm_details->{"OP_NAME"} = "VM_STATUS";

	my $cookie = VRTS::Spoke::get_cookie("VBS");
	my ($ret, $out) = VRTS::Spoke::execute($cs_ip, $cookie, "VBS",
											$vm_details, undef);
	$out =~ s/\n\s*/ /g;

	debug_msg("Response message from VOM: $out");
	if ($out =~ /stderr.*ONLINE/) {
		# TODO: FIXME
		# send update
		$resp->{"code"} = $ret;
		$resp->{"out"} = $out;
	} else {
		$vm_details->{"OP_NAME"} = "VM_START";

		$cookie = VRTS::Spoke::get_cookie("VBS");
		($ret, $out) = VRTS::Spoke::execute($cs_ip, $cookie, "VBS",
											$vm_details, undef);
		$out =~ s/\n\s*/ /g;
		# TODO: FIXME
		# send update
		$resp->{"code"} = $ret;
		# $resp->{"out"} = $out;
	}

	return dispatch_vbs_status($resp, $self);
}

sub stop_vm {
	my ($self) = @_;
	my $resp = {};
	assert_instance($self);

	my $vm_details = {};
	my $vm_name = $self->{"object-name"};
	my $options = $self->{"options"};
	my $cs_ip = $self->{"ConfiguratorIP"};

	$vm_details->{"VM_ID"} = $options->{"vmid"};
	$vm_details->{"SG_ID"} = $options->{"sgid"};
	$vm_details->{"OP_NAME"} = "VM_STOP";

	my $cookie = VRTS::Spoke::get_cookie("VBS");
	my ($ret, $out) = VRTS::Spoke::execute($cs_ip, $cookie, "VBS",
											$vm_details, undef);
	$out =~ s/\n\s*/ /g;

	debug_msg("Response message from VOM: $out");

	# TODO: FIXME
	# send update
	$resp->{"code"} = $ret;
	# $resp->{"out"} = $out;

	return dispatch_vbs_status($resp, $self);
}

sub query {
	my ($self) = @_;
	my $resp = {};
	assert_instance($self);

	my $vm_details = {};
	my $vm_name = $self->{"object-name"};
	my $options = $self->{"options"};
	my $cs_ip = $self->{"ConfiguratorIP"};

	$vm_details->{"VM_ID"} = $options->{"vmid"};
	$vm_details->{"SG_ID"} = $options->{"sgid"};
	$vm_details->{"OP_NAME"} = "VM_STATUS";

	my $cookie = VRTS::Spoke::get_cookie("VBS");
	my ($ret, $out) = VRTS::Spoke::execute($cs_ip, $cookie, "VBS",
											$vm_details, undef);
	$out =~ s/\n\s*/ /g;

	debug_msg("Response message from VOM: $out");
	if ($ret eq 0) {
		# TODO: FIXME
		# send update
		$resp->{"optype"} = "-update";
		$resp->{"objtype"} = "VirtualMachine";
		# TODO: FIXME: make sure vmid is a string
		$resp->{"objname"} = $options->{"vmid"};

		if ($out =~ /stderr.*ONLINE/) {
			$resp->{"VMState"} = 4;
		} elsif ($out =~ /stderr.*OFFLINE/) {
			$resp->{"VMState"} = 2;
		} else {
			$resp->{"VMState"} = 1;
		}
		return dispatch_vbs_status($resp, $self);
	}

	return 0;
}

1;

